using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using MeshVR;
using MVR.FileManagement;
using MVR.FileManagementSecure;

using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Rendering;
using UnityEngine.PostProcessing;

namespace moyashi {  
  public class PostProcessing : MVRScript {
    private static uint VERSION = 7;

    private EffectsManager effectsManager;

    // builtin cameras
    private Camera screenshotCamera;
    private Camera thumbnailCamera;

    private UIDynamicButton customLutButton;
    private UIDynamicTextField customLutTextfield;

    private ImageLoaderThreaded.QueuedImage queuedImage;
    private Texture2D texture;

    List<UIDynamicButton> tabs = new List<UIDynamicButton>();
    List<UIDynamic> colorGradingTab = new List<UIDynamic>();
    List<UIDynamic> lensTab = new List<UIDynamic>();
    List<UIDynamic> atmosphereTab = new List<UIDynamic>();

    private string getPackagePath() {
      string filename = manager.GetJSON()["plugins"][name.Split('_')[0]].Value;
      int index = filename.IndexOf(":/");

      if (index > 0) {
        string path = filename.Substring(0, index + 2);
        // SuperController.LogMessage($"package path: {path}");
        return path;
      } else {
        return "";
      }
    }

    private IEnumerator loadShaders() {
      AssetLoadedMonitor assetLoadedMonitor = new AssetLoadedMonitor();
      AssetLoader.AssetBundleFromFileRequest fileRequest = new AssetLoader.AssetBundleFromFileRequest {
        path = getPackagePath() + $"Custom/Assets/moyashi/Moyashi_PostProcessing{VERSION}.assetbundle",
        callback = assetLoadedMonitor.finish
      };

      // TODO: investigate why we arent calling done enough...
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.QueueLoadAssetBundleFromFile(fileRequest);
      yield return StartCoroutine(assetLoadedMonitor);

      if (fileRequest.assetBundle != null) {
        Shader prepassShader = fileRequest.assetBundle.LoadAsset<Shader>($"MoyashiPrepass{VERSION}.shader");
        Shader colorCorrectionShader = fileRequest.assetBundle.LoadAsset<Shader>($"MoyashiColorCorrection{VERSION}.shader");
        Shader passthroughShader = fileRequest.assetBundle.LoadAsset<Shader>($"Passthrough{VERSION}.shader");
        ComputeShader computeShader = fileRequest.assetBundle.LoadAsset<ComputeShader>($"MoyashiCompute{VERSION}.compute");
        ComputeShader lensComputeShader = fileRequest.assetBundle.LoadAsset<ComputeShader>($"MoyashiLens{VERSION}.compute");

        if (effectsManager != null) {
          effectsManager.setShaders(prepassShader, colorCorrectionShader, computeShader, lensComputeShader);
        } else {
          SuperController.LogError("effects camera was null when trying to load shaders, this shouldnt be possible");
        }
      } else {
        SuperController.LogError("assetbundle failed to load");
      }
    }

    public void Defer(Action f, float delay) {
      StartCoroutine(InvokeRoutine(f, delay));
    }
    
    private IEnumerator InvokeRoutine(System.Action f, float delay) {
      yield return new WaitForSeconds(delay);
      f();
    }

    // applies a setting to all instances of CustomPostProcessing... might change how
    // this works but it's pretty convenient
    private void applySetting(Action<EffectsManager> f) {
      // TODO: rework
      // foreach (CustomPostProcessing customPostProcessing in customPostProcessingSet) {
      //   f(customPostProcessing);
      // }

      f(effectsManager);
    }

    private UIDynamicSlider CreateMiniDynamicSlider(string label, JSONStorableFloat f) {
      UIDynamicSlider slider = CreateSlider(f, false);   
      slider.Configure(label, f.min, f.max, f.val, true, "F3", false, false);   
      slider.defaultButtonEnabled = false;
      slider.quickButtonsGroup.gameObject.SetActive(false);
      // slider.sliderValueTextFromFloat.gameObject.SetActive(false);
      LayoutElement component = slider.GetComponent<LayoutElement>();
      if (component != null) {
        component.minHeight = 0.0f;
        component.preferredHeight = 0.0f;
        component.flexibleHeight = 0.0f;
        component.ignoreLayout = true; // ideally ignore layout, but causes it to spring back to normal height? annoying
        // component.minHeight = 50.0f;
        // component.preferredHeight = 50.0f;
      }

      RectTransform rect = slider.GetComponent<RectTransform>();
      Vector2 offset = rect.offsetMin;
      offset.x = 0.0f;
      offset.y = 0.0f;
      rect.offsetMin = offset;
      rect.sizeDelta = new Vector2(0.0f, 0.0f);

      // turn slider into "mini bar"
      RectTransform rect2 = slider.slider.GetComponent<RectTransform>();
      if (rect2 != null) {
        Vector2 sizeDelta = rect2.sizeDelta;
        // sizeDelta.y = 40.0f;
        sizeDelta.y = 0.0f;
        rect2.sizeDelta = sizeDelta;
        rect2.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0, 430.0f);
        rect2.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 5.0f, 40.0f);

        Vector2 offsetMin = rect2.offsetMin;
        offsetMin.x = 40.0f;
        rect2.offsetMin = offsetMin;
      }

      // make sure the clickable text input doesnt overlap the mini slider bar
      slider.sliderValueTextFromFloat.UIInputField.textComponent.rectTransform.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Right, 0, 85.0f);

      return slider;
    }

    // TODO: probably move this to its own class so we can destruct the custom text elements etc
    private UIDynamic createVectorMixer(JSONStorableVector3 v, bool alignRight) {
      UIDynamic spacer = CreateSpacer(alignRight);
      Transform mixer = spacer.transform;
      // RectTransform rect = mixer.GetComponent<RectTransform>();
      // if (rect != null) {
      //   rect.
      // }
      spacer.height = 178.0f;

      UIContainer container = mixer.gameObject.AddComponent<UIContainer>();

      JSONStorableString storable = new JSONStorableString("_Label", v.name);
      UIDynamicTextField textfield = CreateTextField(storable, alignRight);
      textfield.textColor = new Color(1,1,1);
      textfield.height = 30.0f;
      textfield.UItext.transform.parent = mixer;
      textfield.UItext.fontSize = 24;
      textfield.UItext.fontStyle = FontStyle.Bold;
      textfield.UItext.alignment = TextAnchor.MiddleCenter;

      LayoutElement component = textfield.GetComponent<LayoutElement>();
      component.ignoreLayout = true;

      RectTransform textfieldRect = textfield.GetComponent<RectTransform>();
      textfieldRect.sizeDelta = new Vector2(0.0f, 0.0f);

      RectTransform textRect = textfield.UItext.GetComponent<RectTransform>();
      textRect.localPosition = new Vector3(0,0,0);
      textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 2.5f, 30.0f);
      textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 10.0f, 500.0f);

      // default button
      UIDynamicButton defaultButton = CreateButton("Default", alignRight);
      defaultButton.buttonText.fontSize = 16;

      // dunno why you need to defer all this layout stuff but whatever
      Defer(() => {
        LayoutElement defaultButtonLayout = defaultButton.GetComponent<LayoutElement>();
        defaultButtonLayout.ignoreLayout = true;
        RectTransform defaultButtonRect = defaultButton.GetComponent<RectTransform>();
        defaultButtonRect.parent = container.transform;
        defaultButtonRect.localPosition = new Vector3(0,0,0);
        defaultButtonRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 2.5f, 30.0f);
        defaultButtonRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Right, 2.5f, 100.0f);  
      }, 0.1f);

      // attach and re-align sliders
      JSONStorableFloat r = new JSONStorableFloat(v.name + "Red", v.val[0], v.min[0], v.max[0], true, true);
      RegisterFloat(r);
      { // red        
        UIDynamicSlider sliderR = CreateMiniDynamicSlider("R", r);

        Defer(() => {
          sliderR.transform.parent = mixer;
          LayoutElement layout = sliderR.GetComponent<LayoutElement>();
          layout.ignoreLayout = true;
          RectTransform rect = sliderR.GetComponent<RectTransform>();
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 525.0f); 
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 30.0f, 0.0f);
        }, 0.1f);

        r.setCallbackFunction = (float c) => { v.SetXVal(c); };
      }

      JSONStorableFloat g = new JSONStorableFloat(v.name + "Green", v.val[1], v.min[1], v.max[1], true, true);
      RegisterFloat(g);
      { // green        
        UIDynamicSlider sliderG = CreateMiniDynamicSlider("G", g);

        Defer(() => {
          sliderG.transform.parent = mixer;
          LayoutElement layout = sliderG.GetComponent<LayoutElement>();
          layout.ignoreLayout = true;
          RectTransform rect = sliderG.GetComponent<RectTransform>();
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 525.0f); 
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 77.5f, 0.0f);
        }, 0.1f);

        g.setCallbackFunction = (float c) => { v.SetYVal(c); };
      }

      JSONStorableFloat b = new JSONStorableFloat(v.name + "Blue", v.val[2], v.min[2], v.max[2], true, true);
      RegisterFloat(b);
      { // blue        
        UIDynamicSlider sliderB = CreateMiniDynamicSlider("B", b);

        Defer(() => {
          sliderB.transform.parent = mixer;
          LayoutElement layout = sliderB.GetComponent<LayoutElement>();
          layout.ignoreLayout = true;
          RectTransform rect = sliderB.GetComponent<RectTransform>();
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 525.0f); 
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 125.0f, 0.0f);
        }, 0.1f);

        b.setCallbackFunction = (float c) => { v.SetZVal(c); };
      }

      defaultButton.button.onClick.AddListener(() => {
        r.SetValToDefault();
        g.SetValToDefault();
        b.SetValToDefault();
      });

      return spacer;
    }

    private void initChannelMixer(List<UIDynamic> tab) {
      Color initialR = new Color(1.0f, 0.0f, 0.0f);
      Color initialG = new Color(0.0f, 1.0f, 0.0f);
      Color initialB = new Color(0.0f, 0.0f, 1.0f);

      float min = -1.0f;
      float max = 2.0f;

      Vector3 mixerMin = new Vector3(min, min, min);
      Vector3 mixerMax = new Vector3(max, max, max);

      JSONStorableVector3 mixerR = new JSONStorableVector3("Channel mixer (R)", new Vector3(1.0f, 0.0f, 0.0f), mixerMin, mixerMax, true, true);
      JSONStorableVector3 mixerG = new JSONStorableVector3("Channel mixer (G)", new Vector3(0.0f, 1.0f, 0.0f), mixerMin, mixerMax, true, true);
      JSONStorableVector3 mixerB = new JSONStorableVector3("Channel mixer (B)", new Vector3(0.0f, 0.0f, 1.0f), mixerMin, mixerMax, true, true);

      UIDynamic vectorMixerR = createVectorMixer(mixerR, false);
      UIDynamic vectorMixerG = createVectorMixer(mixerG, false);
      UIDynamic vectorMixerB = createVectorMixer(mixerB, false);

      mixerR.setCallbackFunction = (Vector3 v) => { applySetting((s) => s.setChannelMixR(v)); };
      mixerG.setCallbackFunction = (Vector3 v) => { applySetting((s) => s.setChannelMixG(v)); };
      mixerB.setCallbackFunction = (Vector3 v) => { applySetting((s) => s.setChannelMixB(v)); };

      tab.Add(vectorMixerR);
      tab.Add(vectorMixerG);
      tab.Add(vectorMixerB);
    }

    private void initContrastScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = -1.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Contrast", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setContrastScale(v)); };

      tab.Add(slider);
    }

    private void initTonemappingSelect(List<UIDynamic> tab) {
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 1.0f;

      List<string> tonemappingMethods = new List<string>();
      tonemappingMethods.Add("None");
      tonemappingMethods.Add("ACES");

      JSONStorableStringChooser storableTonemap = new JSONStorableStringChooser("Tonemapper", tonemappingMethods, "ACES", "Tonemap");
      UIDynamicPopup dropdown = CreatePopup(storableTonemap, true);

      JSONStorableFloat storableTonemapStrength = new JSONStorableFloat("Tonemapping strength", initialValue, min, max, true, true);
      storableTonemapStrength.storeType = JSONStorableParam.StoreType.Full;

      storableTonemap.setCallbackFunction = (string s) => {
        switch(s) {
          case "None":
            storableTonemapStrength.val = 0.0f;
            storableTonemapStrength.max = 0.0f;
            break;
          case "ACES":
            storableTonemapStrength.max = max;
            storableTonemapStrength.val = 1.0f;
            break;
        }
      };

      UIDynamicSlider slider = CreateSlider(storableTonemapStrength, true);      
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storableTonemapStrength);
      storableTonemapStrength.setCallbackFunction = (float v) => { applySetting((s) => s.setTonemappingScale(v)); };

      tab.Add(dropdown);
      tab.Add(slider);
    }

    private void initTonemappingExposure(List<UIDynamic> tab) {
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 5.0f;

      JSONStorableFloat storableTonemapExposure = new JSONStorableFloat("Tonemapping exposure", initialValue, min, max, true, true);
      storableTonemapExposure.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storableTonemapExposure, true);      
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storableTonemapExposure);
      storableTonemapExposure.setCallbackFunction = (float v) => { applySetting((s) => s.setTonemappingExposure(v)); };

      tab.Add(slider);
    }

    private void initSaturationScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = -1.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Saturation", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setSaturationScale(v)); };

      tab.Add(slider);
    }

    private void initBloomControl(List<UIDynamic> tab) {
      JSONStorableBool storable = new JSONStorableBool("Bloom", true);
      storable.setCallbackFunction = (bool val) => {
        effectsManager.setBloomEnabled(val);
      };

      RegisterBool(storable);
      var toggle = CreateToggle(storable, false);

      tab.Add(toggle);
    }

    private void initBloomThreshold(List<UIDynamic> tab) {
      float initialValue = 0.75f;
      float min = 0.0f;
      float max = 2.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Bloom Threshold", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setBloomThreshold(v)); };

      tab.Add(slider);
    }

    private void initBloomIntensity(List<UIDynamic> tab) {
      float initialValue = 0.2f;
      float min = 0.0f;
      float max = 2.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Bloom Intensity", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setBloomIntensity(v)); };

      tab.Add(slider);
    }

    private void initBloomRadius(List<UIDynamic> tab) {
      float initialValue = 3.0f;
      float min = 1.0f;
      float max = 7.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Bloom Radius", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setBloomRadius(v)); };

      tab.Add(slider);
    }

    private void initSharpeningScale(List<UIDynamic> tab) {
      float initialValue = 0.3f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Sharpening", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setSharpeningScale(v)); };

      tab.Add(slider);
    }

    private void initVignetteScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Vignette", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setVignetteScale(v)); };

      tab.Add(slider);
    }

    private void initChromaticAberrationScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Chromatic Aberration", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setChromaticAberrationScale(v)); };

      tab.Add(slider);
    }

    private void initDepthOfFieldToggle(List<UIDynamic> tab) {
      JSONStorableBool storable = new JSONStorableBool("Depth of Field", false);
      storable.setCallbackFunction = (bool val) => {
        effectsManager.setDepthOfFieldEnabled(val);
      };

      RegisterBool(storable);
      var toggle = CreateToggle(storable, true);

      tab.Add(toggle);
    }

    private void initFocalLength(List<UIDynamic> tab) {
      float initialValue = 50.0f;
      float min = 1.0f;
      float max = 300.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Focal Length (mm)", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsManager.setDepthOfFieldFocalLength(v); };

      tab.Add(slider);
    }

    private void initApertureSize(List<UIDynamic> tab) {
      float initialValue = 5.6f;
      float min = 0.05f;
      float max = 32.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Aperture (f-stop)", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsManager.setDepthOfFieldAperture(v); };

      tab.Add(slider);
    }

    private void initKernelSize(List<UIDynamic> tab) {
      List<string> kernelSizes = new List<string>();
      kernelSizes.Add("Small");
      kernelSizes.Add("Medium");
      kernelSizes.Add("Large");
      kernelSizes.Add("Very Large");

      JSONStorableStringChooser.SetStringCallback pickerCallback = (string s) => {
        switch(s) {
          case "Small":
            effectsManager.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.Small);
            break;
          case "Medium":
            effectsManager.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.Medium);
            break;
          case "Large":
            effectsManager.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.Large);
            break;
          case "Very Large":
            effectsManager.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.VeryLarge);
            break;
        }
      };
      JSONStorableStringChooser storable = new JSONStorableStringChooser("Kernel Size", kernelSizes, "Medium", "Kernel Size", pickerCallback); // pickerCallback
      UIDynamicPopup dropdown = CreatePopup(storable, true);
      RegisterStringChooser(storable);

      tab.Add(dropdown);
    }

    private void initAutoFocus(List<UIDynamic> tab) {
      float initialValue = 1.5f;
      float min = 0.1f;
      float max = 100.0f;

      UIDynamicPopup targetPopup = null;
      UIDynamicSlider distanceSlider = null;

      List<string> atomList = new List<string>();
      atomList.Add("Manual");

      List<string> targetList = new List<string>();

      JSONStorableStringChooser storableSubTarget = new JSONStorableStringChooser("Target", targetList, "head", "Target");
      RegisterStringChooser(storableSubTarget);

      JSONStorableFloat storableDistance = new JSONStorableFloat("Focus Distance", initialValue, min, max, true, true);
      storableDistance.storeType = JSONStorableParam.StoreType.Full;
      RegisterFloat(storableDistance);

      Dictionary<string, Func<GameObject>> atomDictionary = new Dictionary<string, Func<GameObject>>();
      atomDictionary.Add("Manual", () => null);

      var types = new [] { "Person" };
      var filteredAtoms = SuperController.singleton.GetAtoms().Where(atom => types.Any(atom.type.Equals));

      foreach (var atom in filteredAtoms) {
        atomDictionary.Add(atom.uid, () => null);
        atomList.Add(atom.uid);
      }  

      // OnAtomUIDsChanged maybe?

      JSONStorableStringChooser.SetStringCallback chooserCallback = (string s) => {
        switch(s) {
          case "Manual":
            effectsManager.setAutoFocusTarget(null);
            storableSubTarget.setCallbackFunction = null;

            if (targetPopup != null) {
              tab.Remove(targetPopup);
              RemovePopup(targetPopup);
              targetPopup = null;
            }

            if (distanceSlider == null) {
              distanceSlider = CreateSlider(storableDistance, true);
              distanceSlider.valueFormat = "F3";
              distanceSlider.AutoSetButtons();

              storableDistance.setCallbackFunction = (float v) => { effectsManager.setDepthOfFieldFocusDistance(v); };

              tab.Add(distanceSlider);
            }

            effectsManager.setDepthOfFieldFocusDistance(storableDistance.val);
            break;
          default: 
            if (distanceSlider != null) {
              tab.Remove(distanceSlider);
              RemoveSlider(distanceSlider);
              distanceSlider = null;
            }

            var atom = SuperController.singleton.GetAtomByUid(s);
            var filteredRBs = atom.linkableRigidbodies.Where(rb => atom.forceReceivers.Select(fr => fr.name).Any(rb.name.Equals)); // filter to the non control things
            
            {
              var name = storableSubTarget.val;
              var target = filteredRBs.FirstOrDefault(rb => name == rb.name);
              effectsManager.setAutoFocusTarget(target.gameObject);
            }

            storableSubTarget.choices = filteredRBs.Select((rb) => rb.name).ToList();

            if (targetPopup == null) {
              targetPopup = CreateFilterablePopup(storableSubTarget, true);
              tab.Add(targetPopup);
            }

            storableSubTarget.setCallbackFunction = (name) => {
              var target = filteredRBs.FirstOrDefault(rb => name == rb.name);
              
              if(target != null) {
                effectsManager.setAutoFocusTarget(target.gameObject);
              }
            };

            break;
        }
      };

      JSONStorableStringChooser storable = new JSONStorableStringChooser("Focus", atomList, "Manual", "Focus", chooserCallback); // pickerCallback
      storable.representsAtomUid = true;
      UIDynamicPopup dropdown = CreateFilterablePopup(storable, true);
      RegisterStringChooser(storable);

      if (distanceSlider == null && storable.val == storable.defaultVal) {
        distanceSlider = CreateSlider(storableDistance, true);
        distanceSlider.valueFormat = "F3";
        distanceSlider.AutoSetButtons();

        storableDistance.setCallbackFunction = (float v) => { effectsManager.setDepthOfFieldFocusDistance(v); };

        tab.Add(distanceSlider);
      }

      storable.popupOpenCallback = () => {
        filteredAtoms = SuperController.singleton.GetAtoms().Where(atom => types.Any(atom.type.Equals));

        var newList = new List<string> { "Manual" };
        newList.AddRange(filteredAtoms.Select((atom) => atom.name).ToList());

        storable.choices = newList;
      };


      tab.Add(dropdown);
    }

    private void initSoftenScale(List<UIDynamic> tab) {
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Soften", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setSoftenScale(v)); };

      tab.Add(slider);
    }

    private void initAtmosphereDensity(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Atmosphere Density", initialValue, min, max, false, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.label = "Density";
      slider.valueFormat = "F3";
      slider.rangeAdjustEnabled = true;
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsManager.setAtmosphereDensity(v); };

      tab.Add(slider);
    }

    private void initAtmosphereSeaLevel(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = -10.0f;
      float max = 10.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Atmosphere Sea Level", initialValue, min, max, false, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.label = "Sea Level";
      slider.valueFormat = "F3";
      slider.rangeAdjustEnabled = true;
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsManager.setAtmosphereSeaLevel(v); };

      tab.Add(slider);
    }

    private void initAtmosphereSeaLevelDensity(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Atmosphere Sea Level Density", initialValue, min, max, false, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.label = "Sea Level Density";
      slider.valueFormat = "F3";
      slider.rangeAdjustEnabled = true;
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsManager.setAtmosphereSeaLevelDensity(v); };

      tab.Add(slider);
    }

    private void initCustomLUT(List<UIDynamic> tab) {
      TextureSetCallback callback = TextureLoaded;

      List<string> luts = new List<string>();
      luts.Add("None");
      luts.Add("Moyashi Default");
      luts.Add("Moyashi Energetic");
      luts.Add("Unity Default");
      luts.Add("Custom");

      // built in LUT picker
      JSONStorableUrl storableLutUrl = new JSONStorableUrl("LUT_texture", getPackagePath() + "Custom/Images/LUT/MoyashiDefaultLUT32.png", (string url) => { QueueLoadTexture(url, callback); }, "jpg|png|tif|tiff");
      if (storableLutUrl.val == storableLutUrl.defaultVal) {
        QueueLoadTexture(storableLutUrl.val, callback);
      }
      RegisterUrl(storableLutUrl);

      JSONStorableBool storableCustomSwitch = new JSONStorableBool("LUT_custom_enabled", false);
      RegisterBool(storableCustomSwitch);

      // Check if Custom/Images/LUT/ exists, if not create it.
      if (!FileManagerSecure.DirectoryExists("Custom/Images/LUT", true)) {
        FileManagerSecure.CreateDirectory("Custom/Images/LUT");
      }

      JSONStorableUrl storableLutCustomUrl = new JSONStorableUrl("LUT_custom", string.Empty, (string url) => { 
        if (storableCustomSwitch.val) {
          storableLutUrl.val = url; 
        }
      }, "jpg|png|tif|tiff", "Custom/Images/LUT/");

      List<ShortCut> shortCuts = new List<ShortCut>();

      // hecking booleans, 1st arg: flatten or something? 2nd arg: allow browsing whole pc, 3rd arg include regular dirs
      shortCuts.AddRange(FileManagerSecure.GetShortCutsForDirectory("Custom/Images/LUT/", true, false, true)); 
      
      // GET READY FOR SOME REAL HACKS HOLY
      // basically prefills some actual useful packages in the sidebar rather than mostly poorly packaged stuff
      var creators = new [] { "ICannotDie", "Killer2u", "BooMoon"};
      var directories = new [] { "Custom/LUTs/", "Custom/Assets/ICannotDie/LUT32/", "Custom/Assets/MacGruber/PostMagic/BooMoon_LUT_PACK/" };
      foreach(string directory in directories) {
        var otherShortCuts = FileManagerSecure.GetShortCutsForDirectory(directory);
        if (otherShortCuts.Count > 0) {
          var filteredShortCuts = otherShortCuts.Where(osc => creators.Any(osc.creator.Contains)).ToList();
          shortCuts.AddRange(filteredShortCuts);
        }
      }

      shortCuts.Insert(0, new ShortCut {
        displayName = "Custom\\Images",
        path = FileManagerSecure.NormalizePath("Custom/Images"),
        // flatten = true,
      });

      shortCuts.Insert(0, new ShortCut {
        displayName = "Custom",
        path = FileManagerSecure.NormalizePath("Custom"),
        // flatten = true,
      });

      shortCuts.Insert(0, new ShortCut {
        displayName = "AddonPackages",
        path = FileManagerSecure.NormalizePath("AddonPackages"),
        // flatten = true,
      });

      shortCuts.Insert(0, new ShortCut {
        displayName = "Root",
        path = FileManagerSecure.NormalizePath("."),
        // flatten = true,
      });

      storableLutCustomUrl.beginBrowseWithObjectCallback = (JSONStorableUrl innerStorable) => {
        innerStorable.shortCuts = shortCuts;
      };

      RegisterUrl(storableLutCustomUrl);
      
      JSONStorableStringChooser.SetStringCallback pickerCallback = (string s) => {
        switch(s) {
          case "None":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/NeutralLUT32.png";
            break;
          case "Moyashi Energetic":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/MoyashiEnergeticLUT32.png";
            break;
          case "Moyashi Default":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/MoyashiDefaultLUT32.png";
            break;
          case "Unity Default":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/UnityDefaultLUT32.png";
            break;
          case "Custom":
            storableCustomSwitch.val = true;
            storableLutUrl.val = storableLutCustomUrl.val;
            break;
        }
      };

      JSONStorableStringChooser storable = new JSONStorableStringChooser("LUT_builtin", luts, "Moyashi Default", "LUT", pickerCallback); // pickerCallback
      UIDynamicPopup dropdown = CreatePopup(storable, true);
      RegisterStringChooser(storable);

      // LUT strength
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storableLutStrength = new JSONStorableFloat("LUT strength", initialValue, min, max, true, true);
      storableLutStrength.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storableLutStrength, true);      
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storableLutStrength);
      storableLutStrength.setCallbackFunction = (float v) => { applySetting((s) => s.setLUTContribution(v)); };
      storableCustomSwitch.setCallbackFunction = (bool visible) => {
        if (visible && customLutButton == null && customLutTextfield == null) {
          customLutButton = CreateButton("Browse for Custom LUT", true);
          customLutTextfield = CreateTextField(storableLutCustomUrl, true);
          customLutTextfield.height = 0.0f;
          storableLutCustomUrl.RegisterFileBrowseButton(customLutButton.button);
          tab.Add(customLutButton);
          tab.Add(customLutTextfield);
        } else if (customLutButton != null && customLutTextfield != null) {
          tab.Remove(customLutButton);
          tab.Remove(customLutTextfield);
          RemoveButton(customLutButton);
          RemoveTextField(customLutTextfield);
        }
      };

      tab.Add(dropdown);
      tab.Add(slider);
    }

    private void initSkyColor(List<UIDynamic> tab) {
      Color initialSkyColor = new Color(0.7f, 0.93f, 1.0f);

      HSVColor hsvSkyColor = HSVColorPicker.RGBToHSV(initialSkyColor.r, initialSkyColor.g, initialSkyColor.b);
      JSONStorableColor skyColor = new JSONStorableColor("Sky Color", hsvSkyColor); // should this be JSONStorableVector3?
      skyColor.storeType = JSONStorableParam.StoreType.Full;
      var picker = CreateColorPicker(skyColor, false);
      RegisterColor(skyColor);
      skyColor.setCallbackFunction = (float h, float s, float v) => { effectsManager.setAtmosphereSkyColor(HSVColorPicker.HSVToRGB(h, s, v)); };

      // hack fiesta
      Defer(() => { 
        var rect = picker.colorPicker.Selector.rectTransform;
        var container = picker.colorPicker.Container;

        if (rect != null && container != null && container.rect != null) {
          rect.anchoredPosition = new Vector2(
            255 * picker.colorPicker.xVal,
            container.rect.height * picker.colorPicker.yVal
          );
        }
      }, 0.0f);

      tab.Add(picker);
    }

    private void initSunColor(List<UIDynamic> tab) {
      Color initialSunColor = new Color(1.0f, 0.99f, 0.83f);

      HSVColor hsvSunColor = HSVColorPicker.RGBToHSV(initialSunColor.r, initialSunColor.g, initialSunColor.b);
      JSONStorableColor sunColor = new JSONStorableColor("Sun Color", hsvSunColor); // should this be JSONStorableVector3?
      sunColor.storeType = JSONStorableParam.StoreType.Full;
      var picker = CreateColorPicker(sunColor, false);
      RegisterColor(sunColor);
      sunColor.setCallbackFunction = (float h, float s, float v) => { effectsManager.setAtmosphereSunColor(HSVColorPicker.HSVToRGB(h, s, v)); };

      // hack fiesta
      Defer(() => { 
        var rect = picker.colorPicker.Selector.rectTransform;
        var container = picker.colorPicker.Container;

        if (rect != null && container != null && container.rect != null) {
          rect.anchoredPosition = new Vector2(
            255 * picker.colorPicker.xVal,
            container.rect.height * picker.colorPicker.yVal
          );
        }
      }, 0.0f);

      tab.Add(picker);
    }

    // This runs first when the plugin is loaded, after enable, init gets run
    private void OnEnable() {
      // SuperController.LogMessage("OnEnable");
      Camera mainCamera = CameraTarget.centerTarget.targetCamera;

      // Create a special camera just for doing post processing, although this actually manages
      // effects on multiple cameras
      effectsManager = mainCamera.GetComponent<EffectsManager>();
      if (effectsManager == null) {
        effectsManager = mainCamera.gameObject.AddComponent<EffectsManager>();
      }

      // SuperController.LogMessage("Available Layer masks");

      // for(int i = 0; i < 32; i++) {
        // SuperController.LogMessage($" - {i} {LayerMask.LayerToName(i)}");
      // }
    }

    private void styleButtonAsTab(UIDynamicButton button, bool active) {
      var color = active ? new Color(0.42f, 0.03f, 0.51f) : new Color(0.63f, 0.54f, 0.66f);

      button.buttonColor = color;
      button.buttonText.color = Color.white;
      button.buttonText.alignment = TextAnchor.UpperCenter;
      button.buttonText.fontStyle = FontStyle.Bold;
      {
        var offsetMax = button.buttonText.rectTransform.offsetMax;
        offsetMax.y = -10.5f;
        button.buttonText.rectTransform.offsetMax = offsetMax;
      }
    }

    private void setActive(List<UIDynamic> elements) {
      Defer(() => {
        foreach(var element in elements) {
          element.gameObject.SetActive(true);
        }
      }, 0.0f);
    }

    private void setInactive(List<UIDynamic> elements) {
      Defer(() => {
        foreach(var element in elements) {
          element.gameObject.SetActive(false);
        }
      }, 0.0f);
    }

    private void handleTabClick(UIDynamicButton triggeringTab) {
      foreach(var tab in tabs) {
        styleButtonAsTab(tab, false);
      }

      styleButtonAsTab(triggeringTab, true);

      switch(triggeringTab.label) {
        case "Color Grading":
          setActive(colorGradingTab);
          setInactive(lensTab);
          setInactive(atmosphereTab);
          break;
        case "Lens":
          setInactive(colorGradingTab);
          setActive(lensTab);
          setInactive(atmosphereTab);
          break;
        case "Atmosphere":
          setInactive(colorGradingTab);
          setInactive(lensTab);
          setActive(atmosphereTab);
          break;
      }
    }

    private void initTabs() {
      float height = 40.0f;

      UIDynamic spacerL = CreateSpacer(false);
      spacerL.height = height;

      UIDynamic spacerR = CreateSpacer(true);
      spacerR.height = height;

      UIDynamicButton colorCorrectionTab = CreateButton("Color Grading", false);
      styleButtonAsTab(colorCorrectionTab, true);
      tabs.Add(colorCorrectionTab);
      colorCorrectionTab.button.onClick.AddListener(() => handleTabClick(colorCorrectionTab));

      Defer(() => {
        colorCorrectionTab.transform.parent = spacerL.transform;
        LayoutElement layout = colorCorrectionTab.GetComponent<LayoutElement>();
        layout.ignoreLayout = true;
        RectTransform rect = colorCorrectionTab.GetComponent<RectTransform>();
        rect.sizeDelta = new Vector2(0.0f, 0.0f);
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 210.0f); 
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 0.0f, 60.0f);
      }, 0.1f);

      UIDynamicButton lensTab = CreateButton("Lens", false);
      styleButtonAsTab(lensTab, false);
      tabs.Add(lensTab);
      lensTab.button.onClick.AddListener(() => handleTabClick(lensTab));

      Defer(() => {
        lensTab.transform.parent = spacerL.transform;
        LayoutElement layout = lensTab.GetComponent<LayoutElement>();
        layout.ignoreLayout = true;
        RectTransform rect = lensTab.GetComponent<RectTransform>();
        rect.sizeDelta = new Vector2(0.0f, 0.0f);
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 215.0f, 85.0f); 
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 0.0f, 60.0f);
      }, 0.1f);

      UIDynamicButton atmosphereTab = CreateButton("Atmosphere", false);
      styleButtonAsTab(atmosphereTab, false);
      tabs.Add(atmosphereTab);
      atmosphereTab.button.onClick.AddListener(() => handleTabClick(atmosphereTab));

      Defer(() => {
        atmosphereTab.transform.parent = spacerL.transform;
        LayoutElement layout = atmosphereTab.GetComponent<LayoutElement>();
        layout.ignoreLayout = true;
        RectTransform rect = atmosphereTab.GetComponent<RectTransform>();
        rect.sizeDelta = new Vector2(0.0f, 0.0f);
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 305.0f, 185.0f); 
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 0.0f, 60.0f);
      }, 0.1f);    
    }

    // UIDynamicTextField textfield = CreateTextField(storable, alignRight);
    // textfield.textColor = new Color(1,1,1);
    // textfield.height = 30.0f;
    // textfield.UItext.transform.parent = mixer;
    // textfield.UItext.fontSize = 24;
    // textfield.UItext.fontStyle = FontStyle.Bold;
    // textfield.UItext.alignment = TextAnchor.MiddleCenter;

    // LayoutElement component = textfield.GetComponent<LayoutElement>();
    // component.ignoreLayout = true;

    // RectTransform textfieldRect = textfield.GetComponent<RectTransform>();
    // textfieldRect.sizeDelta = new Vector2(0.0f, 0.0f);

    // RectTransform textRect = textfield.UItext.GetComponent<RectTransform>();
    // textRect.localPosition = new Vector3(0,0,0);
    // textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 2.5f, 30.0f);
    // textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 10.0f, 500.0f);







    // init happens after onenable
    public override void Init() {
      // SuperController.LogMessage("Init");
      // this is roughly in order of what the shader does, whether thats helpful or not

      initTabs();

      initSaturationScale(colorGradingTab);
      initContrastScale(colorGradingTab);
      initChannelMixer(colorGradingTab); 
      initSoftenScale(colorGradingTab);
      initTonemappingSelect(colorGradingTab);
      initTonemappingExposure(colorGradingTab);
      initCustomLUT(colorGradingTab);

      initBloomControl(lensTab);
      initBloomThreshold(lensTab);
      initBloomIntensity(lensTab);
      initBloomRadius(lensTab);
      initSharpeningScale(lensTab);
      initVignetteScale(lensTab);
      initChromaticAberrationScale(lensTab);
      initDepthOfFieldToggle(lensTab);
      initKernelSize(lensTab);
      initApertureSize(lensTab);
      initFocalLength(lensTab);
      initAutoFocus(lensTab);

      initSkyColor(atmosphereTab);
      initSunColor(atmosphereTab);
      initAtmosphereDensity(atmosphereTab);
      initAtmosphereSeaLevelDensity(atmosphereTab);
      initAtmosphereSeaLevel(atmosphereTab);

      handleTabClick(tabs.First());

      effectsManager.enabled = true;

      // TODO: i think initializing the rest of the UI needs to wait for this shader to load first...
      // but eh, just work around it
      StartCoroutine("loadShaders");
    }

    private void OnDisable() {
      // SuperController.LogMessage("main OnDisable");
      
    }

    // This runs when either the plugin is removed or is reloaded
    private void OnDestroy() {
      // SuperController.LogMessage("main OnDisable");
      StopCoroutine("loadShaders");

      if (texture != null) {
        DestroyImmediate(texture);
      }

      texture = null;   

      effectsManager.enabled = false;
      DestroyImmediate(effectsManager);

      // make sure any queued image loads don't run callbacks after destroy
      queuedImage.callback = null;

      AssetLoader.DoneWithAssetBundleFromFile(getPackagePath() + "Custom/Assets/moyashi/Moyashi_CustomColorCorrection.assetbundle");
      // SuperController.LogMessage("main OnDestroy complete");
    }

    private void TextureLoaded(Texture2D tex) {           
      if (texture != null) {
        Destroy(texture);
      }

      if (tex != null) {
        tex.filterMode = FilterMode.Bilinear;
        effectsManager.setLUTTexture(tex);
      }

      queuedImage = null;
    }

    private void QueueLoadTexture(string url, TextureSetCallback callback) { 
      if (ImageLoaderThreaded.singleton == null) {
        SuperController.LogError("no ImageLoaderThreaded, old version of vam?");
        return;
      }

      if (string.IsNullOrEmpty(url)) {
        // This happens when selecting "custom" for the first time...
        // so i guess not really an error just handle it silently
        // SuperController.LogError("null or empty url");
        return;
      }

      // prevent the callbacks stepping on each others toes
      if (queuedImage != null) {
        queuedImage.callback = null;
      }

      queuedImage = new ImageLoaderThreaded.QueuedImage();
      queuedImage.compress = false;
      queuedImage.createMipMaps = false;
      queuedImage.forceReload = true;
      queuedImage.imgPath = url;
      queuedImage.linear = true;
      queuedImage.callback = (ImageLoaderThreaded.QueuedImage queuedImage) => {
        if (queuedImage.tex != null) {
          queuedImage.tex.wrapMode = TextureWrapMode.Clamp;
          queuedImage.tex.filterMode = FilterMode.Bilinear;
          queuedImage.tex.anisoLevel = 0;
          callback(queuedImage.tex);
        }        
      };

      ImageLoaderThreaded.singleton.QueueImage(queuedImage);
    }
  }

  public delegate void TextureSetCallback(Texture2D tex);
}
